@@ -19,6 +19,8 @@
 
 #include "ccache.h"
 
+extern struct conf *conf;
+
 static char *
 find_executable_in_path(const char *name, const char *exclude_name, char *path);
 
@@ -27,8 +29,8 @@ find_executable_in_path(const char *name, const char *exclude_name, char *path);
  * Re-create a win32 command line string based on **argv.
  * http://msdn.microsoft.com/en-us/library/17w5ykft.aspx
  */
-static char *
-argvtos(char *prefix, char **argv)
+char *
+win32argvtos(char *prefix, char **argv)
 {
 	char *arg;
 	char *ptr;
@@ -56,7 +58,7 @@ argvtos(char *prefix, char **argv)
 	} while ((arg = argv[i++]));
 
 	str = ptr = malloc(l + 1);
-	if (str == NULL)
+	if (!str)
 		return NULL;
 
 	i = 0;
@@ -88,22 +90,13 @@ argvtos(char *prefix, char **argv)
 	return str;
 }
 
-int
-win32execute(char *path, char **argv, int doreturn,
-             const char *path_stdout, const char *path_stderr)
+char *
+win32getshell(char *path)
 {
-	PROCESS_INFORMATION pi;
-	STARTUPINFO si;
-	BOOL ret;
-	DWORD exitcode;
 	char *path_env;
 	char *sh = NULL;
-	char *args;
 	const char *ext;
 
-	memset(&pi, 0x00, sizeof(pi));
-	memset(&si, 0x00, sizeof(si));
-
 	ext = get_extension(path);
 	if (ext && strcasecmp(ext, ".sh") == 0 && (path_env = getenv("PATH")))
 		sh = find_executable_in_path("sh.exe", NULL, path_env);
@@ -120,87 +113,150 @@ win32execute(char *path, char **argv, int doreturn,
 			fclose(fp);
 		}
 	}
+
+	return sh;
+}
+
+void add_exe_ext_if_no_to_fullpath(char *full_path_win_ext, size_t max_size,
+                                   const char *ext, const char *path) {
+	if (!ext || (!str_eq(".exe", ext)
+	             && !str_eq(".bat", ext)
+	             && !str_eq(".EXE", ext)
+	             && !str_eq(".BAT", ext))) {
+		snprintf(full_path_win_ext, max_size, "%s.exe", path);
+	} else {
+		snprintf(full_path_win_ext, max_size, "%s", path);
+	}
+}
+
+int
+win32execute(char *path, char **argv, int doreturn,
+             int fd_stdout, int fd_stderr)
+{
+	PROCESS_INFORMATION pi;
+	STARTUPINFO si;
+	BOOL ret;
+	DWORD exitcode;
+	char *sh = NULL;
+	char *args;
+
+	memset(&pi, 0x00, sizeof(pi));
+	memset(&si, 0x00, sizeof(si));
+
+	sh = win32getshell(path);
 	if (sh)
 		path = sh;
 
 	si.cb = sizeof(STARTUPINFO);
-	if (path_stdout) {
-		SECURITY_ATTRIBUTES sa = { sizeof(SECURITY_ATTRIBUTES), NULL, TRUE };
-		si.hStdOutput = CreateFile(path_stdout, GENERIC_WRITE, 0, &sa,
-		                           CREATE_ALWAYS, FILE_ATTRIBUTE_TEMPORARY |
-		                           FILE_FLAG_SEQUENTIAL_SCAN, NULL);
-		si.hStdError  = CreateFile(path_stderr, GENERIC_WRITE, 0, &sa,
-		                           CREATE_ALWAYS, FILE_ATTRIBUTE_TEMPORARY |
-		                           FILE_FLAG_SEQUENTIAL_SCAN, NULL);
-		si.hStdInput  = GetStdHandle(STD_INPUT_HANDLE);
-		si.dwFlags    = STARTF_USESTDHANDLES;
-		if (si.hStdOutput == INVALID_HANDLE_VALUE ||
-		    si.hStdError  == INVALID_HANDLE_VALUE)
+	if (fd_stdout != -1) {
+		si.hStdOutput = (HANDLE)_get_osfhandle(fd_stdout);
+		si.hStdError = (HANDLE)_get_osfhandle(fd_stderr);
+		si.hStdInput = GetStdHandle(STD_INPUT_HANDLE);
+		si.dwFlags = STARTF_USESTDHANDLES;
+		if (si.hStdOutput == INVALID_HANDLE_VALUE
+		    || si.hStdError == INVALID_HANDLE_VALUE) {
 			return -1;
+		}
+	} else {
+		/* redirect subprocess stdout, stderr into current process */
+		si.hStdOutput = GetStdHandle(STD_OUTPUT_HANDLE);
+		si.hStdError = GetStdHandle(STD_ERROR_HANDLE);
+		si.hStdInput = GetStdHandle(STD_INPUT_HANDLE);
+		si.dwFlags = STARTF_USESTDHANDLES;
+		if (si.hStdOutput == INVALID_HANDLE_VALUE
+		    || si.hStdError == INVALID_HANDLE_VALUE) {
+			return -1;
+		}
+	}
+	args = win32argvtos(sh, argv);
+
+	const char *ext = strrchr(path, '.');
+	char full_path_win_ext[MAX_PATH] = {0};
+	add_exe_ext_if_no_to_fullpath(full_path_win_ext, MAX_PATH, ext, path);
+	ret = CreateProcess(full_path_win_ext, args, NULL, NULL, 1, 0, NULL, NULL,
+	                    &si, &pi);
+	if (fd_stdout != -1) {
+		close(fd_stdout);
+		close(fd_stderr);
 	}
-	args = argvtos(sh, argv);
-	ret = CreateProcess(path, args, NULL, NULL, 1, 0, NULL, NULL, &si, &pi);
 	free(args);
-	if (path_stdout) {
-		CloseHandle(si.hStdOutput);
-		CloseHandle(si.hStdError);
-	}
-	if (ret == 0)
+	if (ret == 0) {
+		LPVOID lpMsgBuf;
+		LPVOID lpDisplayBuf;
+		DWORD dw = GetLastError();
+
+		FormatMessage(
+		  FORMAT_MESSAGE_ALLOCATE_BUFFER |
+		  FORMAT_MESSAGE_FROM_SYSTEM |
+		  FORMAT_MESSAGE_IGNORE_INSERTS,
+		  NULL, dw, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR) &lpMsgBuf,
+		  0, NULL);
+
+		lpDisplayBuf =
+		  (LPVOID) LocalAlloc(LMEM_ZEROINIT,
+		                      (lstrlen((LPCTSTR) lpMsgBuf)
+		                       + lstrlen((LPCTSTR) __FILE__) + 200)
+		                      * sizeof(TCHAR));
+		_snprintf((LPTSTR) lpDisplayBuf,
+		          LocalSize(lpDisplayBuf) / sizeof(TCHAR),
+		          TEXT("%s failed with error %d: %s"), __FILE__, dw, lpMsgBuf);
+
+		cc_log("can't execute %s; OS returned error: %s",
+		       full_path_win_ext, (char *)lpDisplayBuf);
+
+		LocalFree(lpMsgBuf);
+		LocalFree(lpDisplayBuf);
+
 		return -1;
+	}
 	WaitForSingleObject(pi.hProcess, INFINITE);
 	GetExitCodeProcess(pi.hProcess, &exitcode);
 	CloseHandle(pi.hProcess);
 	CloseHandle(pi.hThread);
 	if (!doreturn)
-		exit(exitcode);
+		x_exit(exitcode);
 	return exitcode;
 }
 
 #else
 
-/*
-  execute a compiler backend, capturing all output to the given paths
-  the full path to the compiler to run is in argv[0]
-*/
+/* Execute a compiler backend, capturing all output to the given paths the full
+ * path to the compiler to run is in argv[0]. */
 int
-execute(char **argv, const char *path_stdout, const char *path_stderr)
+execute(char **argv, int fd_out, int fd_err, pid_t *pid)
 {
-	pid_t pid;
-	int status, fd_out, fd_err;
+	int status;
 
 	cc_log_argv("Executing ", argv);
 
-	tmp_unlink(path_stdout);
-	fd_out = open(path_stdout, O_WRONLY|O_CREAT|O_TRUNC|O_EXCL|O_BINARY, 0666);
-	if (fd_out == -1) {
-		fatal("Error creating %s: %s", path_stdout, strerror(errno));
+	block_signals();
+	*pid = fork();
+	unblock_signals();
+
+	if (*pid == -1) {
+		fatal("Failed to fork: %s", strerror(errno));
 	}
 
-	tmp_unlink(path_stderr);
-	fd_err = open(path_stderr, O_WRONLY|O_CREAT|O_TRUNC|O_EXCL|O_BINARY, 0666);
-	if (fd_err == -1) {
-		fatal("Error creating %s: %s", path_stderr, strerror(errno));
-	}
-
-	pid = fork();
-	if (pid == -1) fatal("Failed to fork: %s", strerror(errno));
-
-	if (pid == 0) {
+	if (*pid == 0) {
 		/* Child. */
 		dup2(fd_out, 1);
 		close(fd_out);
 		dup2(fd_err, 2);
 		close(fd_err);
-		exit(execv(argv[0], argv));
+		x_exit(execv(argv[0], argv));
 	}
 
 	close(fd_out);
 	close(fd_err);
 
-	if (waitpid(pid, &status, 0) != pid) {
+	if (waitpid(*pid, &status, 0) != *pid) {
 		fatal("waitpid failed: %s", strerror(errno));
 	}
 
+	block_signals();
+	*pid = 0;
+	unblock_signals();
+
 	if (WEXITSTATUS(status) == 0 && WIFSIGNALED(status)) {
 		return -1;
 	}
@@ -213,7 +269,7 @@ execute(char **argv, const char *path_stdout, const char *path_stderr)
 /*
  * Find an executable by name in $PATH. Exclude any that are links to
  * exclude_name.
-*/
+ */
 char *
 find_executable(const char *name, const char *exclude_name)
 {
@@ -223,8 +279,8 @@ find_executable(const char *name, const char *exclude_name)
 		return x_strdup(name);
 	}
 
-	path = getenv("CCACHE_PATH");
-	if (!path) {
+	path = conf->path;
+	if (str_eq(path, "")) {
 		path = getenv("PATH");
 	}
 	if (!path) {
@@ -249,11 +305,14 @@ find_executable_in_path(const char *name, const char *exclude_name, char *path)
 	     tok = strtok_r(NULL, PATH_DELIM, &saveptr)) {
 #ifdef _WIN32
 		char namebuf[MAX_PATH];
-		int ret = SearchPath(tok, name, ".exe",
+		int ret = SearchPath(tok, name, NULL,
 		                     sizeof(namebuf), namebuf, NULL);
-		if (!ret)
-			ret = SearchPath(tok, name, NULL,
+		if (!ret) {
+			char *exename = format("%s.exe", name);
+			ret = SearchPath(tok, exename, NULL,
 			                 sizeof(namebuf), namebuf, NULL);
+			free(exename);
+		}
 		(void) exclude_name;
 		if (ret) {
 			free(path);
