@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2010-2011 Joel Rosdahl
+ * Copyright (C) 2010-2015 Joel Rosdahl
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the Free
@@ -21,9 +21,12 @@
  */
 
 #include "ccache.h"
+#include "conf.h"
 #include "test/framework.h"
 #include "test/util.h"
 
+extern struct conf *conf;
+
 TEST_SUITE(argument_processing)
 
 TEST(dash_E_should_result_in_called_for_preprocessing)
@@ -33,7 +36,7 @@ TEST(dash_E_should_result_in_called_for_preprocessing)
 
 	create_file("foo.c", "");
 	CHECK(!cc_process_args(orig, &preprocessed, &compiler));
-	CHECK_UNS_EQ(1, stats_get_pending(STATS_PREPROCESSING));
+	CHECK_INT_EQ(1, stats_get_pending(STATS_PREPROCESSING));
 
 	args_free(orig);
 }
@@ -45,7 +48,7 @@ TEST(dash_M_should_be_unsupported)
 
 	create_file("foo.c", "");
 	CHECK(!cc_process_args(orig, &preprocessed, &compiler));
-	CHECK_UNS_EQ(1, stats_get_pending(STATS_UNSUPPORTED));
+	CHECK_INT_EQ(1, stats_get_pending(STATS_UNSUPPORTED));
 
 	args_free(orig);
 }
@@ -53,9 +56,32 @@ TEST(dash_M_should_be_unsupported)
 TEST(dependency_flags_should_only_be_sent_to_the_preprocessor)
 {
 #define CMD \
-	"cc -c -MD -MMD -MP -MF foo.d -MT mt1 -MT mt2 -MQ mq1 -MQ mq2" \
+	"cc -MD -MMD -MP -MF foo.d -MT mt1 -MT mt2 -MQ mq1 -MQ mq2" \
 	" -Wp,-MD,wpmd -Wp,-MMD,wpmmd"
-	struct args *orig = args_init_from_string(CMD " foo.c -o foo.o");
+	struct args *orig = args_init_from_string(CMD " -c foo.c -o foo.o");
+	struct args *exp_cpp = args_init_from_string(CMD);
+#undef CMD
+	struct args *exp_cc = args_init_from_string("cc -c");
+	struct args *act_cpp = NULL, *act_cc = NULL;
+	create_file("foo.c", "");
+
+	CHECK(cc_process_args(orig, &act_cpp, &act_cc));
+	CHECK_ARGS_EQ_FREE12(exp_cpp, act_cpp);
+	CHECK_ARGS_EQ_FREE12(exp_cc, act_cc);
+
+	args_free(orig);
+}
+
+TEST(preprocessor_only_flags_should_only_be_sent_to_the_preprocessor)
+{
+#define CMD \
+	"cc -I. -idirafter . -iframework. -imacros . -imultilib ." \
+	" -include test.h -include-pch test.pch -iprefix . -iquote ." \
+	" -isysroot . -isystem . -iwithprefix . -iwithprefixbefore ." \
+	" -DTEST_MACRO -DTEST_MACRO2=1 -F. -trigraphs -fworking-directory" \
+	" -fno-working-directory -MD -MMD -MP -MF foo.d -MT mt1 -MT mt2 " \
+	" -MQ mq1 -MQ mq2 -Wp,-MD,wpmd -Wp,-MMD,wpmmd"
+	struct args *orig = args_init_from_string(CMD " -c foo.c -o foo.o");
 	struct args *exp_cpp = args_init_from_string(CMD);
 #undef CMD
 	struct args *exp_cc = args_init_from_string("cc -c");
@@ -72,9 +98,9 @@ TEST(dependency_flags_should_only_be_sent_to_the_preprocessor)
 TEST(dependency_flags_that_take_an_argument_should_not_require_space_delimiter)
 {
 	struct args *orig = args_init_from_string(
-		"cc -c -MMD -MFfoo.d -MT mt -MTmt -MQmq foo.c -o foo.o");
+	  "cc -c -MMD -MFfoo.d -MT mt -MTmt -MQmq foo.c -o foo.o");
 	struct args *exp_cpp = args_init_from_string(
-		"cc -c -MMD -MFfoo.d -MT mt -MTmt -MQmq");
+	  "cc -MMD -MFfoo.d -MT mt -MTmt -MQmq");
 	struct args *exp_cc = args_init_from_string("cc -c");
 	struct args *act_cpp = NULL, *act_cc = NULL;
 	create_file("foo.c", "");
@@ -88,34 +114,32 @@ TEST(dependency_flags_that_take_an_argument_should_not_require_space_delimiter)
 
 TEST(sysroot_should_be_rewritten_if_basedir_is_used)
 {
-	extern char *base_dir;
 	extern char *current_working_dir;
-	struct args *orig =
-		args_init_from_string("cc --sysroot=/some/directory -c foo.c");
+	char *arg_string;
+	struct args *orig;
 	struct args *act_cpp = NULL, *act_cc = NULL;
+
 	create_file("foo.c", "");
-
-	CHECK(cc_process_args(orig, &act_cpp, &act_cc));
-	CHECK_STR_EQ(act_cpp->argv[1], "--sysroot=/some/directory");
-
-	cc_reset();
-	base_dir = "/some";
+	free(conf->base_dir);
+	conf->base_dir = x_strdup("/");
 	current_working_dir = get_cwd();
+	arg_string = format("cc --sysroot=%s/foo -c foo.c", current_working_dir);
+	orig = args_init_from_string(arg_string);
 
 	CHECK(cc_process_args(orig, &act_cpp, &act_cc));
-	CHECK(str_startswith(act_cpp->argv[1], "--sysroot=../"));
+	CHECK_STR_EQ(act_cpp->argv[1], "--sysroot=./foo");
 
 	args_free(orig);
-	base_dir = NULL;
-	current_working_dir = NULL;
+	args_free(act_cpp);
+	args_free(act_cc);
 }
 
 TEST(MF_flag_with_immediate_argument_should_work_as_last_argument)
 {
 	struct args *orig = args_init_from_string(
-		"cc -c foo.c -o foo.o -MMD -MT bar -MFfoo.d");
+	  "cc -c foo.c -o foo.o -MMD -MT bar -MFfoo.d");
 	struct args *exp_cpp = args_init_from_string(
-		"cc -c -MMD -MT bar -MFfoo.d");
+	  "cc -MMD -MT bar -MFfoo.d");
 	struct args *exp_cc = args_init_from_string("cc -c");
 	struct args *act_cpp = NULL, *act_cc = NULL;
 	create_file("foo.c", "");
@@ -130,9 +154,9 @@ TEST(MF_flag_with_immediate_argument_should_work_as_last_argument)
 TEST(MT_flag_with_immediate_argument_should_work_as_last_argument)
 {
 	struct args *orig = args_init_from_string(
-		"cc -c foo.c -o foo.o -MMD -MFfoo.d -MT foo -MTbar");
+	  "cc -c foo.c -o foo.o -MMD -MFfoo.d -MT foo -MTbar");
 	struct args *exp_cpp = args_init_from_string(
-		"cc -c -MMD -MFfoo.d -MT foo -MTbar");
+	  "cc -MMD -MFfoo.d -MT foo -MTbar");
 	struct args *exp_cc = args_init_from_string("cc -c");
 	struct args *act_cpp = NULL, *act_cc = NULL;
 	create_file("foo.c", "");
@@ -147,9 +171,9 @@ TEST(MT_flag_with_immediate_argument_should_work_as_last_argument)
 TEST(MQ_flag_with_immediate_argument_should_work_as_last_argument)
 {
 	struct args *orig = args_init_from_string(
-		"cc -c foo.c -o foo.o -MMD -MFfoo.d -MQ foo -MQbar");
+	  "cc -c foo.c -o foo.o -MMD -MFfoo.d -MQ foo -MQbar");
 	struct args *exp_cpp = args_init_from_string(
-		"cc -c -MMD -MFfoo.d -MQ foo -MQbar");
+	  "cc -MMD -MFfoo.d -MQ foo -MQbar");
 	struct args *exp_cc = args_init_from_string("cc -c");
 	struct args *act_cpp = NULL, *act_cc = NULL;
 	create_file("foo.c", "");
@@ -164,11 +188,10 @@ TEST(MQ_flag_with_immediate_argument_should_work_as_last_argument)
 TEST(MQ_flag_without_immediate_argument_should_not_add_MQobj)
 {
 	struct args *orig = args_init_from_string(
-		"gcc -c -MD -MP -MFfoo.d -MQ foo.d foo.c");
+	  "gcc -c -MD -MP -MFfoo.d -MQ foo.d foo.c");
 	struct args *exp_cpp = args_init_from_string(
-		"gcc -c -MD -MP -MFfoo.d -MQ foo.d");
-	struct args *exp_cc = args_init_from_string(
-		"gcc -c");
+	  "gcc -MD -MP -MFfoo.d -MQ foo.d");
+	struct args *exp_cc = args_init_from_string("gcc -c");
 	struct args *act_cpp = NULL, *act_cc = NULL;
 	create_file("foo.c", "");
 
@@ -182,11 +205,10 @@ TEST(MQ_flag_without_immediate_argument_should_not_add_MQobj)
 TEST(MT_flag_without_immediate_argument_should_not_add_MTobj)
 {
 	struct args *orig = args_init_from_string(
-		"gcc -c -MD -MP -MFfoo.d -MT foo.d foo.c");
+	  "gcc -c -MD -MP -MFfoo.d -MT foo.d foo.c");
 	struct args *exp_cpp = args_init_from_string(
-		"gcc -c -MD -MP -MFfoo.d -MT foo.d");
-	struct args *exp_cc = args_init_from_string(
-		"gcc -c");
+	  "gcc -MD -MP -MFfoo.d -MT foo.d");
+	struct args *exp_cc = args_init_from_string("gcc -c");
 	struct args *act_cpp = NULL, *act_cc = NULL;
 	create_file("foo.c", "");
 
@@ -197,14 +219,13 @@ TEST(MT_flag_without_immediate_argument_should_not_add_MTobj)
 	args_free(orig);
 }
 
-TEST(MQ_flag_with_immediate_argument_should_add_MQobj)
+TEST(MQ_flag_with_immediate_argument_should_not_add_MQobj)
 {
 	struct args *orig = args_init_from_string(
-		"gcc -c -MD -MP -MFfoo.d -MQfoo.d foo.c");
+	  "gcc -c -MD -MP -MFfoo.d -MQfoo.d foo.c");
 	struct args *exp_cpp = args_init_from_string(
-		"gcc -c -MD -MP -MFfoo.d -MQfoo.d -MQ foo.o");
-	struct args *exp_cc = args_init_from_string(
-		"gcc -c");
+	  "gcc -MD -MP -MFfoo.d -MQfoo.d");
+	struct args *exp_cc = args_init_from_string("gcc -c");
 	struct args *act_cpp = NULL, *act_cc = NULL;
 	create_file("foo.c", "");
 
@@ -215,14 +236,13 @@ TEST(MQ_flag_with_immediate_argument_should_add_MQobj)
 	args_free(orig);
 }
 
-TEST(MT_flag_with_immediate_argument_should_add_MQobj)
+TEST(MT_flag_with_immediate_argument_should_not_add_MQobj)
 {
 	struct args *orig = args_init_from_string(
-		"gcc -c -MD -MP -MFfoo.d -MTfoo.d foo.c");
+	  "gcc -c -MD -MP -MFfoo.d -MTfoo.d foo.c");
 	struct args *exp_cpp = args_init_from_string(
-		"gcc -c -MD -MP -MFfoo.d -MTfoo.d -MQ foo.o");
-	struct args *exp_cc = args_init_from_string(
-		"gcc -c");
+	  "gcc -MD -MP -MFfoo.d -MTfoo.d");
+	struct args *exp_cc = args_init_from_string("gcc -c");
 	struct args *act_cpp = NULL, *act_cc = NULL;
 	create_file("foo.c", "");
 
@@ -233,5 +253,48 @@ TEST(MT_flag_with_immediate_argument_should_add_MQobj)
 	args_free(orig);
 }
 
+TEST(fprofile_flag_with_existing_dir_should_be_rewritten_to_real_path)
+{
+	struct args *orig = args_init_from_string(
+	  "gcc -c -fprofile-generate=some/dir foo.c");
+	struct args *exp_cpp = args_init_from_string("gcc");
+	struct args *exp_cc = args_init_from_string("gcc");
+	struct args *act_cpp = NULL, *act_cc = NULL;
+	char *s;
+
+	create_file("foo.c", "");
+	mkdir("some", 0777);
+	mkdir("some/dir", 0777);
+	s = format("-fprofile-generate=%s", x_realpath("some/dir"));
+	args_add(exp_cpp, s);
+	args_add(exp_cc, s);
+	args_add(exp_cc, "-c");
+	free(s);
+
+	CHECK(cc_process_args(orig, &act_cpp, &act_cc));
+	CHECK_ARGS_EQ_FREE12(exp_cpp, act_cpp);
+	CHECK_ARGS_EQ_FREE12(exp_cc, act_cc);
+
+	args_free(orig);
+}
+
+TEST(fprofile_flag_with_nonexisting_dir_should_not_be_rewritten)
+{
+	struct args *orig = args_init_from_string(
+	  "gcc -c -fprofile-generate=some/dir foo.c");
+	struct args *exp_cpp = args_init_from_string(
+	  "gcc -fprofile-generate=some/dir");
+	struct args *exp_cc = args_init_from_string(
+	  "gcc -fprofile-generate=some/dir -c");
+	struct args *act_cpp = NULL, *act_cc = NULL;
+
+	create_file("foo.c", "");
+
+	CHECK(cc_process_args(orig, &act_cpp, &act_cc));
+	CHECK_ARGS_EQ_FREE12(exp_cpp, act_cpp);
+	CHECK_ARGS_EQ_FREE12(exp_cc, act_cc);
+
+	args_free(orig);
+}
 
 TEST_SUITE_END
